# -*- coding: utf-8 -*-

import os
import socket
import thread
import time
import zmq
from ..share import constants, shine_pb2
from ..share.log import logger


class Connection(object):

    work_progress = None

    address_list = None
    zmq_client = None

    def __init__(self, app, address_list):
        self.app = app
        self.address_list = address_list

        ctx = zmq.Context()
        self.zmq_client = ctx.socket(zmq.PULL)
        for address in address_list:
            self.zmq_client.connect(address)

    def run(self):
        thread.start_new_thread(self._monitor_work_timeout, ())
        while 1:
            try:
                self._handle()
            except KeyboardInterrupt:
                break
            except:
                logger.error('exc occur.', exc_info=True)

    def _monitor_work_timeout(self):
        """
        监控work的耗时
        :return:
        """

        while self.app.enable:
            time.sleep(1)

            work_progress = self.work_progress
            if work_progress:
                past_time = time.time() - work_progress['begin_time']
                if self.app.work_timeout is not None and past_time > self.app.work_timeout:
                    # 说明worker的处理时间已经太长了
                    logger.error('work timeout: %s / %s, request: %s',
                                 past_time, self.app.work_timeout, work_progress['request'])
                    # 强制从子线程退出worker
                    os._exit(-1)

    def _handle(self):
        if not self.app.enable:
            # 安全退出
            raise KeyboardInterrupt

        self._read_message()

    def write(self, data):
        """
        格式可以支持到frame
        发送数据    True: 成功   else: 失败
        """
        if self.closed():
            logger.error('connection closed. data: %r', data)
            return False

        # 只支持字符串
        self.app.events.before_response(self, data)
        for bp in self.app.blueprints:
            bp.events.before_app_response(self, data)

        self.app.forwarder_client.send(data)

        for bp in self.app.blueprints:
            bp.events.after_app_response(self, data)
        self.app.events.after_response(self, data)

        return True

    def _read_message(self):

        data = self.zmq_client.recv()

        self._on_read_complete(data)

        if self.closed():
            self._on_connection_close()

    def _on_connection_close(self):
        # 链接被关闭的回调

        logger.error('connection closed, address_list: %s', self.address_list)

        for bp in self.app.blueprints:
            bp.events.close_app_conn(self)
        self.app.events.close_conn(self)

    def _on_read_complete(self, data):
        """
        数据获取结束
        """
        task = shine_pb2.Task()
        task.ParseFromString(data)

        logger.debug('task:\n%s', task)

        request = self.app.request_class(self, task)

        # 设置task开始处理的时间和信息
        self.work_progress = dict(
            begin_time=time.time(),
            request=request,
        )
        self._handle_request(request)
        self.work_progress = None

    def _handle_request(self, request):
        """
        出现任何异常的时候，服务器不再主动关闭连接
        """

        if not request.is_valid:
            return False

        if request.task.cmd == constants.CMD_CLIENT_CREATED:
            self.app.events.create_client(request)
            for bp in self.app.blueprints:
                bp.events.create_app_client(request)
            return True
        elif request.task.cmd == constants.CMD_CLIENT_CLOSED:
            self.app.events.close_client(request)
            for bp in self.app.blueprints:
                bp.events.close_app_client(request)
            return True

        if not request.view_func:
            logger.info('cmd invalid. request: %s' % request)
            if not request.responded:
                request.write_to_client(dict(ret=constants.RET_INVALID_CMD))
            return False

        if not self.app.got_first_request:
            self.app.got_first_request = True
            self.app.events.before_first_request(request)
            for bp in self.app.blueprints:
                bp.events.before_app_first_request(request)

        self.app.events.before_request(request)
        for bp in self.app.blueprints:
            bp.events.before_app_request(request)
        if request.blueprint:
            request.blueprint.events.before_request(request)

        if request.interrupted:
            # 业务要求中断
            return True

        view_func_exc = None

        try:
            request.view_func(request)
        except Exception, e:
            logger.error('view_func raise exception. request: %s, e: %s',
                         request, e, exc_info=True)
            view_func_exc = e
            # 必须是没有回应过
            if not request.responded:
                request.write_to_client(dict(ret=constants.RET_INTERNAL))

        if request.blueprint:
            request.blueprint.events.after_request(request, view_func_exc)
        for bp in self.app.blueprints:
            bp.events.after_app_request(request, view_func_exc)
        self.app.events.after_request(request, view_func_exc)

        return True

    def close(self):
        """
        直接关闭连接
        """
        self.zmq_client.close()

    def closed(self):
        """
        连接是否已经关闭
        :return:
        """
        return self.zmq_client.closed




# -*- coding: utf-8 -*-

__version__ = '0.1.15'

from gateway.gateway import Gateway
from forwarder.forwarder import Forwarder
from worker.worker import Worker
from worker.blueprint import Blueprint
from trigger.trigger import Trigger
from share.log import logger


# -*- coding: utf-8 -*-

import uuid

import gevent

from ...share.log import logger


class Connection(object):

    # 唯一ID
    id = None

    # 登录之后的用户ID
    uid = 0

    # 用户userdata
    userdata = 0

    def __init__(self, app, stream, address):
        self.id = uuid.uuid4().bytes
        self.app = app
        self.stream = stream
        self.address = address
        self.app.events.create_conn(self)

    def write(self, data):
        """
        发送数据    True: 成功   else: 失败
        """
        if self.stream.closed():
            return False

        if isinstance(data, self.app.box_class):
            # 打包
            data = data.pack()
        elif isinstance(data, dict):
            data = self.app.box_class(data).pack()

        ret = self.stream.write(data)

        return ret

    def close(self, exc_info=False):
        """
        直接关闭连接
        """
        self.stream.close(exc_info)

    def closed(self):
        """
        连接是否已经关闭
        :return:
        """
        return self.stream.closed()

    def handle(self):
        """
        启动处理
        """
        # while中判断可以保证connection_close事件只触发一次
        while not self.stream.closed():
            # 防止内存泄露
            job = gevent.spawn(self._read_message)
            job.join()

    def _read_message(self):
        data = self.stream.read_with_checker(self.app.stream_checker)
        if data:
            self._on_read_complete(data)

        # 在这里加上判断，因为如果在处理函数里关闭了conn，会导致无法触发on_connction_close
        if self.stream.closed():
            self._on_connection_close()

    def _on_connection_close(self):
        # 链接被关闭的回调

        self.app.events.close_conn(self)

    def _on_read_complete(self, data):
        """
        数据获取结束
        """
        try:
            self.app.events.handle_request(self, data)
        except Exception, e:
            logger.error('view_func raise exception. data: %s, e: %s',
                         data, e, exc_info=True)

    def __repr__(self):
        return str(self.id)



# -*- coding: utf-8 -*-

from ..share.shine_pb2 import Task, RspToUsers, CloseUsers
from ..share import constants


class Trigger(object):

    box_class = None

    zmq_client = None

    def __init__(self, box_class, forwarder_input_address_list=None, zmq_client=None, use_gevent=False):

        assert not (forwarder_input_address_list is zmq_client is None)

        self.box_class = box_class

        if forwarder_input_address_list is not None:
            if use_gevent:
                import zmq.green as zmq  # for gevent
            else:
                import zmq

            ctx = zmq.Context()
            self.zmq_client = ctx.socket(zmq.PUSH)
            for address in forwarder_input_address_list:
                self.zmq_client.connect(address)
        else:
            self.zmq_client = zmq_client

    def write_to_users(self, data_list):
        """
        格式为
        [(uids, box), (uids, box, userdata) ...]
        :param data_list: userdata可不传，默认为0，conn.userdata & userdata == userdata
        :return:
        """

        msg = RspToUsers()

        for data_tuple in data_list:
            if len(data_tuple) == 2:
                uids, data = data_tuple
                userdata = None
            else:
                uids, data, userdata = data_tuple

            if isinstance(data, self.box_class):
                data = data.pack()
            elif isinstance(data, dict):
                data = self.box_class(data).pack()

            row = msg.rows.add()
            row.buf = data
            row.userdata = userdata or 0
            row.uids.extend(uids)

        task = Task()
        task.cmd = constants.CMD_WRITE_TO_USERS
        task.data = msg.SerializeToString()

        return self.zmq_client.send(task.SerializeToString())

    def close_users(self, uids, userdata=None):
        msg = CloseUsers()
        msg.uids.extend(uids)
        msg.userdata = userdata or 0

        task = Task()
        task.cmd = constants.CMD_CLOSE_USERS
        task.data = msg.SerializeToString()

        return self.zmq_client.send(task.SerializeToString())

    def write_to_worker(self, data, node_id):
        """
        透传到worker进行处理
        还是得把node_id传过来，告诉forwarder，是要哪个node_id进行处理
        """

        task = Task()
        task.node_id = node_id
        task.cmd = constants.CMD_WRITE_TO_WORKER

        if isinstance(data, self.box_class):
            # 打包
            data = data.pack()
        elif isinstance(data, dict):
            data = self.box_class(data).pack()

        task.data = data

        return self.zmq_client.send(task.SerializeToString())



# -*- coding: utf-8 -*-

import functools
from events import Events

from .utils import safe_func


def _reg_event_handler(func):
    @functools.wraps(func)
    def func_wrapper(obj, handler):
        event = getattr(obj.events, func.__name__)
        event += safe_func(handler)

        return handler
    return func_wrapper


class AppEventsMixin(object):
    events = None

    def __init__(self):
        self.events = Events()

    @_reg_event_handler
    def create_conn(self, f):
        """
        连接建立成功后
        f(conn)
        """

    @_reg_event_handler
    def handle_request(self, f):
        """
        处理request
        f(conn, data)
        """

    @_reg_event_handler
    def close_conn(self, f):
        """
        连接close之后
        f(conn)
        """



# -*- coding: utf-8 -*-

from ..share.log import logger
from ..share import constants
from ..share.shine_pb2 import RspToUsers, CloseUsers, Task
from ..trigger.trigger import Trigger


class Request(object):
    """
    请求
    """

    # 与gateway的连接对象
    conn = None
    # gateway封装的task
    task = None
    # 业务box
    box = None
    # 数据是否有效
    is_valid = False
    # 是否已经作出回应
    responded = False
    blueprint = None
    # 路由表
    route_rule = None
    # 是否中断处理，即不调用view_func，主要用在before_request中
    interrupted = False

    trigger = None

    def __init__(self, conn, task):
        self.conn = conn
        self.task = task
        # 赋值
        self.is_valid = self._parse_raw_data()
        self.trigger = Trigger(self.app.box_class, zmq_client=self.app.forwarder_client)

    @property
    def app(self):
        return self.conn.app

    def _parse_raw_data(self):
        if not self.task.data:
            return True

        try:
            self.box = self.app.box_class()
        except Exception, e:
            logger.error('parse raw_data fail. e: %s, request: %s', e, self)
            return False

        if self.box.unpack(self.task.data) > 0:
            self._parse_route_rule()
            return True
        else:
            logger.error('unpack fail. request: %s', self)
            return False

    def _parse_route_rule(self):
        if self.cmd is None:
            return

        route_rule = self.app.get_route_rule(self.cmd)
        if route_rule:
            # 在app层，直接返回
            self.route_rule = route_rule
            return

        for bp in self.app.blueprints:
            route_rule = bp.get_route_rule(self.cmd)
            if route_rule:
                self.blueprint = bp
                self.route_rule = route_rule
                break

    @property
    def cmd(self):
        try:
            return self.box.cmd
        except:
            return None

    @property
    def view_func(self):
        return self.route_rule['view_func'] if self.route_rule else None

    @property
    def endpoint(self):
        if not self.route_rule:
            return None

        bp_endpoint = self.route_rule['endpoint']

        return '.'.join([self.blueprint.name, bp_endpoint] if self.blueprint else [bp_endpoint])

    def write_to_client(self, data):
        """
        写回
        :param data: 可以是dict也可以是box
        :return:
        """

        assert not (self.app.rsp_once and self.responded), 'request has been responded'

        if isinstance(data, self.app.box_class):
            data = data.pack()
        elif isinstance(data, dict):
            data = self.box.map(data).pack()

        task = Task()
        # 就可以直接通过node_id和client_id来进行识别了
        task.client_id = self.task.client_id
        task.node_id = self.task.node_id
        task.cmd = constants.CMD_WRITE_TO_CLIENT
        task.data = data

        succ = self.conn.write(task.SerializeToString())

        if succ:
            # 如果发送成功，就标记为已经回应
            self.responded = True

        return succ

    def close_client(self):
        task = Task()
        task.client_id = self.task.client_id
        task.node_id = self.task.node_id
        task.cmd = constants.CMD_CLOSE_CLIENT

        return self.conn.write(task.SerializeToString())

    def login_client(self, uid, userdata=None):

        task = Task()
        task.client_id = self.task.client_id
        task.node_id = self.task.node_id
        task.cmd = constants.CMD_LOGIN_CLIENT
        task.uid = uid
        task.userdata = userdata or 0

        return self.conn.write(task.SerializeToString())

    def logout_client(self):
        task = Task()
        task.client_id = self.task.client_id
        task.node_id = self.task.node_id
        task.cmd = constants.CMD_LOGOUT_CLIENT

        return self.conn.write(task.SerializeToString())

    def write_to_users(self, data_list):
        """
        格式为
        [(uids, box), (uids, box, userdata) ...]
        :param data_list: userdata可不传，默认为0，conn.userdata & userdata == userdata
        :return:
        """

        return self.trigger.write_to_users(data_list)

    def close_users(self, uids, userdata=None):
        return self.trigger.close_users(uids, userdata)

    def write_to_worker(self, data):
        """
        透传到worker进行处理
        """

        return self.trigger.write_to_worker(data, node_id=self.task.node_id)

    def interrupt(self, data=None):
        """
        中断处理
        :param data: 要响应的数据，不传即不响应
        :return:
        """
        self.interrupted = True
        if data is not None:
            return self.write_to_client(data)
        else:
            return True

    def __repr__(self):
        return 'cmd: %r, endpoint: %s, task:\n%s' % (self.cmd, self.endpoint, self.task)



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    name="shine",
    version='0.1.15',
    zip_safe=False,
    platforms='any',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    scripts=['shine/bin/shine'],
    install_requires=['events', 'pyzmq', 'netkit', 'click', 'setproctitle'],
    url="https://github.com/dantezhu/shine",
    license="MIT",
    author="dantezhu",
    author_email="zny2008@gmail.com",
    description="shine",
)



# -*- coding: utf-8 -*-

"""
共享存储
用来保存进程唯一ID，用户所在位置等
"""


class ShareStore(object):

    rds = None
    user_key_prefix = None
    nodes_key = None
    user_maxage = None

    def __init__(self, rds, user_key_prefix, nodes_key, user_maxage=None):
        """
        :param rds: redis实例
        :param user_key_prefix: 用户数据的key前缀
        :param nodes_key: 进程集合的key
        :return:
        """

        self.rds = rds
        self.user_key_prefix = user_key_prefix
        self.nodes_key = nodes_key
        self.user_maxage = user_maxage

    def add_user(self, uid, node_id):
        """
        添加用户
        :param uid:
        :param node_id:
        :return:
        """
        return self.rds.set(self._make_redis_key(uid), node_id, ex=self.user_maxage)

    def remove_user(self, uid, node_id):
        """
        删除用户，要保证node_id相等
        :param uid:
        :return:
        """

        old_node_id = self.rds.get(self._make_redis_key(uid))

        if old_node_id is None:
            # 没数据当然直接返回啦
            return

        return self.rds.delete(self._make_redis_key(uid))

    def get_users(self, uid_list):
        """
        批量获取用户信息 {
            1: "223",
            2: "333",
        }
        :param uid_list:
        :return:
        """
        node_id_list = self.rds.mget([self._make_redis_key(uid) for uid in uid_list])

        return dict(zip(uid_list, node_id_list))

    def add_node(self, node_id):
        """
        添加node_id
        :return:
        """

        return self.rds.sadd(self.nodes_key, node_id)

    def remove_node(self, node_id):
        """
        删除node_id
        :return:
        """

        return self.rds.srem(self.nodes_key, node_id)

    def get_nodes(self):
        """
        获取proc集合
        :return:
        """
        return self.rds.smembers(self.nodes_key)

    def _make_redis_key(self, uid):
        return self.user_key_prefix + str(uid)



# -*- coding: utf-8 -*-


import signal
import sys
import weakref
import uuid
import setproctitle

import gevent
from gevent.queue import Queue
import zmq.green as zmq  # for gevent

from .server import Server
from ..share.proc_mgr import ProcMgr
from ..share.log import logger
from ..share import constants, shine_pb2
from ..share.config import ConfigAttribute, Config
from ..share.share_store import ShareStore
from ..share.utils import import_module_or_string


class Gateway(object):
    ############################## configurable begin ##############################

    box_class = ConfigAttribute('BOX_CLASS',
                                get_converter=import_module_or_string)
    name = ConfigAttribute('NAME')
    debug = ConfigAttribute('DEBUG')

    ############################## configurable end   ##############################

    config = None

    proc_mgr = None
    outer_server = None
    inner_server = None
    forwarder_client = None

    # 准备发送到worker的queue
    task_queue = None

    # worker唯一标示
    node_id = None
    # 连接ID->conn
    conn_dict = None
    # 用户ID->conn
    user_dict = None

    # 共享存储
    share_store = None

    def __init__(self):
        self.config = Config(defaults=constants.DEFAULT_CONFIG)
        self.proc_mgr = ProcMgr()
        self.outer_server = Server(self.box_class, self.config['GATEWAY_BACKLOG'])
        self.task_queue = Queue()
        self.conn_dict = dict()
        self.user_dict = weakref.WeakValueDictionary()

    def run(self, debug=None):
        """
        启动
        :param debug: 是否debug
        :return:
        """

        if debug is not None:
            self.debug = debug

        if self.config['REDIS_URL']:
            import redis
            rds = redis.from_url(self.config['REDIS_URL'])
            self.share_store = ShareStore(rds,
                                          self.config['REDIS_USER_KEY_PREFIX'],
                                          self.config['REDIS_NODES_KEY'],
                                          self.config['REDIS_USER_MAXAGE'],
                                          )

        workers = len(self.config['GATEWAY_INNER_ADDRESS_LIST'])

        def run_wrapper():
            logger.info('Running server, debug: %s, workers: %s',
                        self.debug, workers)

            self._prepare_server()
            setproctitle.setproctitle(self._make_proc_name('gateway:master'))
            # 只能在主线程里面设置signals
            self._handle_parent_proc_signals()
            self.proc_mgr.fork_workers(workers, self._worker_run)

        run_wrapper()

    def _make_proc_name(self, subtitle):
        """
        获取进程名称
        :param subtitle:
        :return:
        """
        proc_name = '[%s %s:%s] %s' % (
            self.name,
            constants.NAME,
            subtitle,
            ' '.join([sys.executable] + sys.argv)
        )

        return proc_name

    def _make_redis_key(self, uid):
        return self.config['REDIS_USER_KEY_PREFIX'] + str(uid)

    def _prepare_server(self):
        """
        准备server，因为fork之后就晚了
        :return:
        """
        self.outer_server._prepare_server((self.config['GATEWAY_OUTER_HOST'], self.config['GATEWAY_OUTER_PORT']))

    def _serve_forever(self):
        """
        保持运行
        :return:
        """
        job_list = []
        for action in [self.outer_server._serve_forever, self._fetch_from_forwarder, self._send_task_to_worker]:
            job = gevent.spawn(action)
            job_list.append(job)

        for job in job_list:
            job.join()

    def _start_inner_server(self, address):
        """
        zmq的内部server
        每个worker绑定的地址都要不一样
        """
        ctx = zmq.Context()
        self.inner_server = ctx.socket(zmq.PUSH)
        self.inner_server.bind(address)

    def _fetch_from_forwarder(self):
        """
        从forwarder server那拿数据
        :return:
        """

        ctx = zmq.Context()
        self.forwarder_client = ctx.socket(zmq.SUB)
        for address in self.config['FORWARDER_OUTPUT_ADDRESS_LIST']:
            self.forwarder_client.connect(address)
        self.forwarder_client.setsockopt(zmq.SUBSCRIBE, self.node_id)

        while True:
            topic, msg = self.forwarder_client.recv_multipart()

            task = shine_pb2.Task()
            task.ParseFromString(msg)

            logger.debug('task:\n%s', task)

            # 这样就不会内存泄露了
            job = gevent.spawn(self._handle_task, task)
            job.join()

    def _handle_task(self, task):
        """
        处理task
        :param task:
        :return:
        """

        if task.cmd == constants.CMD_WRITE_TO_CLIENT:
            conn = self.conn_dict.get(task.client_id)
            if conn:
                conn.write(task.data)
        elif task.cmd == constants.CMD_WRITE_TO_WORKER:
            # 重新转发处理
            # 标记一下
            task.inner = 1
            self.task_queue.put(task)
        elif task.cmd == constants.CMD_CLOSE_CLIENT:
            conn = self.conn_dict.get(task.client_id)
            if conn:
                conn.close()
        elif task.cmd == constants.CMD_LOGIN_CLIENT:
            conn = self.conn_dict.get(task.client_id)
            if conn:
                if conn.uid is not None:
                    # 旧的登录用户

                    # 先从存储删掉
                    if self.share_store:
                        self.share_store.remove_user(conn.uid, self.node_id)

                    self.user_dict.pop(conn.uid, None)
                    conn.uid = conn.userdata = 0

                conn.uid = task.uid
                conn.userdata = task.userdata
                self.user_dict[conn.uid] = conn

                # 后写入存储
                if self.share_store:
                    self.share_store.add_user(conn.uid, self.node_id)

        elif task.cmd == constants.CMD_LOGOUT_CLIENT:
            conn = self.conn_dict.get(task.client_id)
            if conn:
                if conn.uid is not None:
                    if self.share_store:
                        self.share_store.remove_user(conn.uid)

                    self.user_dict.pop(conn.uid, None)
                    conn.uid = conn.userdata = 0

        elif task.cmd == constants.CMD_WRITE_TO_USERS:
            rsp = shine_pb2.RspToUsers()
            rsp.ParseFromString(task.data)

            for row in rsp.rows:
                # -1: 所有已登录连接
                # -2: 所有连接
                # -3: 所有未登陆连接
                if -1 in row.uids:
                    for conn in self.conn_dict.values():
                        if conn and conn.uid and (conn.userdata & row.userdata) == row.userdata:
                            conn.write(row.buf)
                elif -2 in row.uids:
                    for conn in self.conn_dict.values():
                        if conn and (conn.userdata & row.userdata) == row.userdata:
                            conn.write(row.buf)
                elif -3 in row.uids:
                    for conn in self.conn_dict.values():
                        if conn and not conn.uid and (conn.userdata & row.userdata) == row.userdata:
                            conn.write(row.buf)
                else:
                    for uid in row.uids:
                        conn = self.user_dict.get(uid)
                        if conn and (conn.userdata & row.userdata) == row.userdata:
                            conn.write(row.buf)

        elif task.cmd == constants.CMD_CLOSE_USERS:
            rsp = shine_pb2.CloseUsers()
            rsp.ParseFromString(task.data)

            if -1 in rsp.uids:
                for conn in self.conn_dict.values():
                    if conn and conn.uid and (conn.userdata & rsp.userdata) == rsp.userdata:
                        conn.close()
            elif -2 in rsp.uids:
                for conn in self.conn_dict.values():
                    if conn and (conn.userdata & rsp.userdata) == rsp.userdata:
                        conn.close()
            elif -3 in rsp.uids:
                for conn in self.conn_dict.values():
                    if conn and not conn.uid and (conn.userdata & rsp.userdata) == rsp.userdata:
                        conn.close()
            else:
                for uid in rsp.uids:
                    conn = self.user_dict.get(uid)
                    if conn and (conn.userdata & rsp.userdata) == rsp.userdata:
                        conn.close()

    def _send_task_to_worker(self):
        """
        将任务发送到worker
        :return:
        """

        while True:
            task = self.task_queue.get()
            self.inner_server.send(task.SerializeToString())

    def _register_outer_server_handlers(self):
        """
        注册server的一些回调
        :return:
        """

        @self.outer_server.create_conn
        def create_conn(conn):
            logger.debug('conn.id: %r', conn.id)
            self.conn_dict[conn.id] = conn

            task = shine_pb2.Task()
            task.node_id = self.node_id
            task.client_id = conn.id
            task.client_ip = conn.address[0]
            task.cmd = constants.CMD_CLIENT_CREATED

            self.task_queue.put(task)

        @self.outer_server.close_conn
        def close_conn(conn):
            # 删除
            logger.debug('conn.id: %r', conn.id)
            self.conn_dict.pop(conn.id, None)
            if conn.uid is not None:
                if self.share_store:
                    self.share_store.remove_user(conn.uid, self.node_id)

                self.user_dict.pop(conn.uid, None)
                conn.uid = conn.userdata = 0

            task = shine_pb2.Task()
            task.node_id = self.node_id
            task.client_id = conn.id
            task.client_ip = conn.address[0]
            task.cmd = constants.CMD_CLIENT_CLOSED

            self.task_queue.put(task)

        @self.outer_server.handle_request
        def handle_request(conn, data):
            # 转发到worker
            logger.debug('conn.id: %r, data: %r', conn.id, data)
            task = shine_pb2.Task()
            task.node_id = self.node_id
            task.client_id = conn.id
            task.client_ip = conn.address[0]
            task.cmd = constants.CMD_CLIENT_REQ
            task.data = data
            self.task_queue.put(task)

    def _worker_run(self, index):
        """
        在worker里面执行的
        :return:
        """
        setproctitle.setproctitle(self._make_proc_name('gateway:app:%s' % index))
        self.node_id = uuid.uuid4().bytes
        if self.share_store:
            self.share_store.add_node(self.node_id)
        self._handle_child_proc_signals()
        self._register_outer_server_handlers()
        self._start_inner_server(self.config['GATEWAY_INNER_ADDRESS_LIST'][index])

        try:
            self._serve_forever()
        except KeyboardInterrupt:
            if self.share_store:
                self.share_store.remove_node(self.node_id)
            pass
        except:
            logger.error('exc occur.', exc_info=True)

    def _handle_parent_proc_signals(self):
        def exit_handler(signum, frame):
            self.proc_mgr.enable = False

            # 如果是终端直接CTRL-C，子进程自然会在父进程之后收到INT信号，不需要再写代码发送
            # 如果直接kill -INT $parent_pid，子进程不会自动收到INT
            # 所以这里可能会导致重复发送的问题，重复发送会导致一些子进程异常，所以在子进程内部有做重复处理判断。
            self.proc_mgr.terminate_all()

        # INT, QUIT, TERM为强制结束
        signal.signal(signal.SIGINT, exit_handler)
        signal.signal(signal.SIGQUIT, exit_handler)
        signal.signal(signal.SIGTERM, exit_handler)

    def _handle_child_proc_signals(self):
        def exit_handler(signum, frame):
            # 防止重复处理KeyboardInterrupt，导致抛出异常
            if self.proc_mgr.enable:
                self.proc_mgr.enable = False
                raise KeyboardInterrupt

        # 强制结束，抛出异常终止程序进行
        signal.signal(signal.SIGINT, exit_handler)
        signal.signal(signal.SIGQUIT, exit_handler)
        signal.signal(signal.SIGTERM, exit_handler)
